from django.db import models


class NotificationsConfigManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related("notifications_api_service")



from celery.exceptions import Ignore, Retry
from celery.utils.time import get_exponential_backoff_interval
from vine.utils import wraps

from .models import NotificationsConfig


def add_autoretry_behaviour(task, **options):
    """
    Adapted from celery to use admin configurable autoretry settings
    celery original code: https://github.com/celery/celery/blob/master/celery/app/autoretry.py

    Why we implement it this way:
    autoretry in Celery is implemented using `add_autoretry_behaviour` which basically
    patches `task.run` method during task initialization.
    `add_autoretry_behaviour` evaluates all the autoretry properties which can be specified
    in @celery.task decorator or in the `Task` class. Either way they are specified and evaluated
    during task initialization, which happens only once for each task.

    That means, that it is impossible to add dynamic autoretry properties without changing
    `add_autoretry_behaviour` behaviour: even if we add descriptors or class properties
    they would be evaluated only once.

    The only solution to add dynamic properties is to patch `task.run` method, so it will evaluate
    autoretry options during calling of the task.

    The purpose of this long docstring is to share the pain and to warn future developers who would
    want to optimize autoretry implementation
    """
    autoretry_for = tuple(
        options.get("autoretry_for", getattr(task, "autoretry_for", ()))
    )
    retry_kwargs = options.get("retry_kwargs", getattr(task, "retry_kwargs", {}))
    retry_jitter = options.get("retry_jitter", getattr(task, "retry_jitter", True))

    if autoretry_for and not hasattr(task, "_orig_run"):

        @wraps(task.run)
        def run(*args, **kwargs):
            config = NotificationsConfig.get_solo()
            max_retries = config.notification_delivery_max_retries
            retry_backoff = config.notification_delivery_retry_backoff
            retry_backoff_max = config.notification_delivery_retry_backoff_max

            task.max_retries = max_retries

            try:
                return task._orig_run(*args, **kwargs)
            except Ignore:
                # If Ignore signal occurs task shouldn't be retried,
                # even if it suits autoretry_for list
                raise
            except Retry:
                raise
            except autoretry_for as exc:
                if retry_backoff:
                    retry_kwargs["countdown"] = get_exponential_backoff_interval(
                        factor=retry_backoff,
                        retries=task.request.retries,
                        maximum=retry_backoff_max,
                        full_jitter=retry_jitter,
                    )
                # Override max_retries
                if hasattr(task, "override_max_retries"):
                    retry_kwargs["max_retries"] = getattr(
                        task, "override_max_retries", max_retries
                    )

                ret = task.retry(exc=exc, **retry_kwargs)
                # Stop propagation
                if hasattr(task, "override_max_retries"):
                    delattr(task, "override_max_retries")
                raise ret

        task._orig_run, task.run = task.run, run



"""
Provide notifications kanaal/exchange classes.
"""
from collections import defaultdict
from typing import Dict, Tuple

from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db.models import Model
from django.db.models.base import ModelBase

KANAAL_REGISTRY = set()


class Kanaal:
    def __init__(self, label: str, main_resource: ModelBase, kenmerken: Tuple = None):
        self.label = label
        self.main_resource = main_resource

        self.usage = defaultdict(list)  # filled in by metaclass of notifications

        # check that we're refering to existing fields
        self.kenmerken = kenmerken or ()
        for kenmerk in self.kenmerken:
            try:
                self.main_resource._meta.get_field(kenmerk)
            except FieldDoesNotExist as exc:
                raise ImproperlyConfigured(
                    f"Kenmerk '{kenmerk}' does not exist on the model {main_resource}"
                ) from exc

        KANAAL_REGISTRY.add(self)

    def __repr__(self):
        cls_name = self.__class__.__name__
        return "%s(label=%r, main_resource=%r)" % (
            cls_name,
            self.label,
            self.main_resource,
        )

    def get_kenmerken(self, obj: Model, data: Dict = None) -> Dict:
        data = data or {}
        return {
            kenmerk: data.get(kenmerk, getattr(obj, kenmerk))
            for kenmerk in self.kenmerken
        }

    def get_usage(self):
        return self.usage.items()

    @property
    def description(self):
        kenmerk_template = "* `{kenmerk}`: {help_text}"
        kenmerken = [
            kenmerk_template.format(
                kenmerk=kenmerk,
                help_text=self.main_resource._meta.get_field(kenmerk).help_text,
            )
            for kenmerk in self.kenmerken
        ]

        description = (
            "**Main resource**\n\n"
            "`{options.model_name}`\n\n\n\n"
            "**Kenmerken**\n\n"
            "{kenmerken}"
        ).format(options=self.main_resource._meta, kenmerken="\n".join(kenmerken))

        return description



# Generated by Django 3.2.14 on 2022-07-11 15:36

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("notifications_api_common", "0003_migrate_api_root"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="notificationsconfig",
            name="api_root",
        ),
    ]



import logging

from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand
from django.urls import reverse

from ...kanalen import KANAAL_REGISTRY
from ...models import NotificationsConfig
from ...settings import get_setting

logger = logging.getLogger(__name__)


class KanaalExists(Exception):
    pass


def create_kanaal(kanaal: str) -> None:
    """
    Create a kanaal, if it doesn't exist yet.
    """
    client = NotificationsConfig.get_client()

    # look up the exchange in the registry
    _kanaal = next(k for k in KANAAL_REGISTRY if k.label == kanaal)

    kanalen = client.list("kanaal", query_params={"naam": kanaal})
    if kanalen:
        raise KanaalExists()

    # build up own documentation URL
    domain = Site.objects.get_current().domain
    protocol = "https" if get_setting("IS_HTTPS") else "http"
    documentation_url = (
        f"{protocol}://{domain}{reverse('notifications:kanalen')}#{kanaal}"
    )

    client.create(
        "kanaal",
        {
            "naam": kanaal,
            "documentatieLink": documentation_url,
            "filters": list(_kanaal.kenmerken),
        },
    )


class Command(BaseCommand):
    help = "Create kanaal in notification component"

    def add_arguments(self, parser):
        parser.add_argument(
            "--kanalen",
            nargs="*",
            type=str,
            help=(
                "Names of the kanalen (will use the `KANAAL_REGISTRY` if not specified)"
            ),
        )

    def handle(self, **options):
        config = NotificationsConfig.get_solo()

        if not config.notifications_api_service:
            self.stderr.write(
                "NotificationsConfig does not have a "
                "`notifications_api_service` configured"
            )

        api_root = config.notifications_api_service.api_root

        # use CLI arg or fall back to setting
        kanalen = options["kanalen"] or sorted(
            [kanaal.label for kanaal in KANAAL_REGISTRY]
        )

        for kanaal in kanalen:
            try:
                create_kanaal(kanaal)
                self.stdout.write(f"Registered kanaal '{kanaal}' with {api_root}")
            except KanaalExists:
                self.stderr.write(f"Kanaal '{kanaal}' already exists within {api_root}")



import logging

from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

logger = logging.getLogger(__name__)


@deconstructible
class UntilNowValidator:
    """
    Validate a datetime to not be in the future.

    This means that `now` is included.
    """

    message = _("Ensure this value is not in the future.")
    code = "future_not_allowed"

    @property
    def limit_value(self):
        return timezone.now()

    def __call__(self, value):
        if value > self.limit_value:
            raise ValidationError(self.message, code=self.code)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.message == other.message
            and self.code == other.code
        )



from django.conf import settings
from django.db import models
from django.http import HttpRequest
from django.urls import Resolver404, ResolverMatch, get_resolver, get_script_prefix

from rest_framework.viewsets import ViewSet

from .kanalen import Kanaal


def notification_documentation(kanaal: Kanaal):
    """
    Generate notification documentation for an OpenAPI specification containing
    the relevant resources and actions for a given KANAAL
    """
    doc = f"""Deze API publiceert notificaties op het kanaal `{kanaal.label}`.

{kanaal.description}

**Resources en acties**
"""
    for resource, actions in kanaal.get_usage():
        doc += f"""- `{resource}`: {', '.join(actions)}\n"""
    return doc


class NotAViewSet(Exception):
    pass


def resolve_path(path: str, resolver=None, script_prefix=None) -> ResolverMatch:
    resolver = resolver or get_resolver()
    prefix = script_prefix or get_script_prefix()
    path = path.replace(prefix, "/", 1)
    try:
        return resolver.resolve(path)
    except Resolver404 as exc:
        raise models.ObjectDoesNotExist("URL did not resolve") from exc


def get_viewset_for_path(path: str, method="GET") -> "ViewSet":
    """
    Look up which viewset matches a path.
    """
    # NOTE: this doesn't support setting a different urlconf on the request
    callback, callback_args, callback_kwargs = resolve_path(path)

    if not hasattr(callback, "cls"):
        raise NotAViewSet(f"Callback for {path} does not look like a viewset")

    viewset = callback.cls(**callback.initkwargs)
    viewset.action_map = callback.actions
    viewset.request = HttpRequest()
    viewset.args = callback_args
    viewset.kwargs = callback_kwargs

    viewset.action = viewset.action_map.get(method.lower())

    return viewset


def get_resource_for_path(path: str) -> models.Model:
    """
    Retrieve the API instance belonging to a (detail) path.
    """
    if settings.FORCE_SCRIPT_NAME and path.startswith(settings.FORCE_SCRIPT_NAME):
        prefix_length = len(settings.FORCE_SCRIPT_NAME)
        path = path[prefix_length:]

    viewset = get_viewset_for_path(path)

    # See rest_framework.mixins.RetieveModelMixin.get_object()
    lookup_url_kwarg = viewset.lookup_url_kwarg or viewset.lookup_field
    filter_kwargs = {viewset.lookup_field: viewset.kwargs[lookup_url_kwarg]}

    return viewset.get_queryset().get(**filter_kwargs)



# Generated by Django 3.2.15 on 2022-10-18 14:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("notifications_api_common", "0005_auto_20220810_0950"),
    ]

    operations = [
        migrations.AddField(
            model_name="notificationsconfig",
            name="notification_delivery_max_retries",
            field=models.PositiveIntegerField(
                default=5,
                help_text="the maximum number of automatic retries. After this amount of retries, guaranteed delivery stops trying to deliver the message.",
            ),
        ),
        migrations.AddField(
            model_name="notificationsconfig",
            name="notification_delivery_retry_backoff",
            field=models.PositiveIntegerField(
                default=3,
                help_text="if specified, a factor applied to the exponential backoff. This allows you to tune how quickly automatic retries are performed.",
            ),
        ),
        migrations.AddField(
            model_name="notificationsconfig",
            name="notification_delivery_retry_backoff_max",
            field=models.PositiveIntegerField(
                default=48,
                help_text="an upper limit in seconds to the exponential backoff time.",
            ),
        ),
    ]



# Generated by Django 3.2.14 on 2022-08-10 09:50

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("zgw_consumers", "0015_auto_20220307_1522"),
        ("notifications_api_common", "0004_remove_notificationsconfig_api_root"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="subscription",
            name="config",
        ),
        migrations.AlterField(
            model_name="notificationsconfig",
            name="notifications_api_service",
            field=models.ForeignKey(
                blank=True,
                limit_choices_to={"api_type": "nrc"},
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="zgw_consumers.service",
                verbose_name="notifications api service",
            ),
        ),
    ]



from django.contrib import admin, messages
from django.utils.translation import gettext_lazy as _

from requests.exceptions import RequestException
from solo.admin import SingletonModelAdmin
from zds_client.client import ClientError

from .models import NotificationsConfig, Subscription


@admin.register(NotificationsConfig)
class NotificationsConfigAdmin(SingletonModelAdmin):
    pass


def register_webhook(modeladmin, request, queryset):
    for sub in queryset:
        if sub._subscription:
            continue

        try:
            sub.register()
        except (ClientError, RequestException) as e:
            messages.error(
                request,
                _(
                    "Something went wrong while registering subscription "
                    "for {callback}: {exception}"
                ).format(callback=sub.callback_url, exception=e),
            )


register_webhook.short_description = _("Register the webhooks")  # noqa


@admin.register(Subscription)
class SubscriptionAdmin(admin.ModelAdmin):
    list_display = ("callback_url", "channels", "_subscription")
    actions = [register_webhook]



from typing import Optional

from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import gettext_lazy as _

from solo.models import SingletonModel
from zds_client import Client, ClientAuth
from zgw_consumers.constants import APITypes
from zgw_consumers.models import Service

from .query import NotificationsConfigManager


class NotificationsConfig(SingletonModel):
    notifications_api_service = models.ForeignKey(
        Service,
        limit_choices_to={"api_type": APITypes.nrc},
        verbose_name=_("notifications api service"),
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
    )
    notification_delivery_max_retries = models.PositiveIntegerField(
        help_text=_(
            "The maximum number of automatic retries. After this amount of retries, "
            "guaranteed delivery stops trying to deliver the message."
        ),
        default=5,
    )
    notification_delivery_retry_backoff = models.PositiveIntegerField(
        help_text=_(
            "If specified, a factor applied to the exponential backoff. "
            "This allows you to tune how quickly automatic retries are performed."
        ),
        default=3,
    )
    notification_delivery_retry_backoff_max = models.PositiveIntegerField(
        help_text=_("An upper limit in seconds to the exponential backoff time."),
        default=48,
    )

    objects = NotificationsConfigManager()

    class Meta:
        verbose_name = _("Notificatiescomponentconfiguratie")

    def __str__(self):
        api_root = (
            self.notifications_api_service.api_root
            if self.notifications_api_service
            else _("no service configured")
        )
        return _("Notifications API configuration ({api_root})").format(
            api_root=api_root
        )

    @classmethod
    def get_client(cls) -> Optional[Client]:
        """
        Construct a client, prepared with the required auth.
        """
        config = cls.get_solo()
        if config.notifications_api_service:
            return config.notifications_api_service.build_client()
        return None


class Subscription(models.Model):
    """
    A single subscription.

    TODO: on change/update, update the subscription
    """

    callback_url = models.URLField(
        _("callback url"), help_text=_("Where to send the notifications (webhook url)")
    )
    client_id = models.CharField(
        _("client ID"),
        max_length=50,
        help_text=_("Client ID to construct the auth token"),
    )
    secret = models.CharField(
        _("client secret"),
        max_length=50,
        help_text=_("Secret to construct the auth token"),
    )
    channels = ArrayField(
        models.CharField(max_length=100),
        verbose_name=_("channels"),
        help_text=_("Comma-separated list of channels to subscribe to"),
    )

    _subscription = models.URLField(
        _("NC subscription"),
        blank=True,
        editable=False,
        help_text=_("Subscription as it is known in the NC"),
    )

    class Meta:
        verbose_name = _("Webhook subscription")
        verbose_name_plural = _("Webhook subscriptions")

    def __str__(self):
        return f"{', '.join(self.channels)} - {self.callback_url}"

    def register(self) -> None:
        """
        Registers the webhook with the notification component.
        """
        assert (
            NotificationsConfig.get_solo().notifications_api_service
        ), "No service for Notifications API configured"

        client = NotificationsConfig.get_client()

        # This authentication is for the NC to call us. Thus, it's *not* for
        # calling the NC to create a subscription.
        # TODO should be replaced with `TokenAuth`
        # see: maykinmedia/notifications-api-common/pull/1#discussion_r941450384
        self_auth = ClientAuth(
            client_id=self.client_id,
            secret=self.secret,
        )
        data = {
            "callbackUrl": self.callback_url,
            "auth": self_auth.credentials()["Authorization"],
            "kanalen": [
                {
                    "naam": channel,
                    # FIXME: You need to be able to configure these.
                    "filters": {},
                }
                for channel in self.channels
            ],
        }

        # register the subscriber
        subscriber = client.create("abonnement", data=data)

        self._subscription = subscriber["url"]
        self.save(update_fields=["_subscription"])



# Generated by Django 3.2.12 on 2022-12-14 06:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("notifications_api_common", "0006_auto_20221213_0214"),
        ("notifications_api_common", "0007_auto_20221206_0414"),
    ]

    operations = []



# Generated by Django 3.2.14 on 2022-07-11 15:10

from django.db import migrations

from zgw_consumers.constants import APITypes
from zgw_consumers.service import Service as _Service


def migrate_notifications_config_forwards(apps, _):
    NotificationsConfig = apps.get_model("notifications_api_common.NotificationsConfig")
    Service = apps.get_model("zgw_consumers.Service")

    config, _ = NotificationsConfig.objects.get_or_create()

    direct_match = _Service.get_service(config.api_root)
    if direct_match:
        config.notifications_api_service = Service.objects.get(pk=direct_match.pk)
    else:
        service = Service.objects.filter(api_type=APITypes.nrc).first()
        config.notifications_api_service = service
    config.save()


def migrate_notifications_config_backwards(apps, _):
    NotificationsConfig = apps.get_model("notifications_api_common.NotificationsConfig")

    config, _ = NotificationsConfig.objects.get_or_create()
    config.api_root = getattr(config.notifications_api_service, "api_root", "")


class Migration(migrations.Migration):

    dependencies = [
        (
            "notifications_api_common",
            "0002_notificationsconfig_notifications_api_service",
        ),
    ]

    operations = [
        migrations.RunPython(
            migrate_notifications_config_forwards,
            migrate_notifications_config_backwards,
        )
    ]



import logging
from contextlib import contextmanager
from typing import Dict, List, Union
from urllib.parse import urlparse

from django.core.exceptions import ImproperlyConfigured
from django.db import models, transaction
from django.utils import timezone

from djangorestframework_camel_case.util import camelize
from rest_framework.permissions import SAFE_METHODS
from rest_framework.routers import SimpleRouter

from .api.serializers import NotificatieSerializer
from .kanalen import Kanaal
from .models import NotificationsConfig
from .settings import get_setting
from .tasks import send_notification
from .utils import get_resource_for_path, get_viewset_for_path

logger = logging.getLogger(__name__)


@contextmanager
def _fake_atomic():
    yield


def conditional_atomic(wrap: bool = True):
    """
    Wrap either a fake or real atomic transaction context manager.
    """
    return transaction.atomic if wrap else _fake_atomic


class NotificationMixinBase(type):
    def __new__(cls, name, bases, attrs):
        new_cls = super().__new__(cls, name, bases, attrs)

        kanaal = attrs.get("notifications_kanaal")
        if kanaal is None:
            return new_cls

        relevant_bases = [base for base in bases if issubclass(base, NotificationMixin)]
        resource = new_cls.queryset.model._meta.model_name
        # use the router to figure out which actions are available
        router = SimpleRouter()
        for route in router.get_routes(new_cls):
            for method, action in route.mapping.items():

                if method.upper() in SAFE_METHODS:
                    continue

                # check if the action is actually provided by one of the mixins
                if not any(hasattr(base, action) for base in relevant_bases):
                    continue

                kanaal.usage[resource].append(action)

        return new_cls


class NotificationMixin(metaclass=NotificationMixinBase):
    notifications_kanaal = None  # must be set be subclasses
    notifications_main_resource_key = None
    notifications_wrap_in_atomic_block = True

    def get_kanaal(self):
        if not self.notifications_kanaal:
            raise ImproperlyConfigured(
                "'%s' should either include a `notifications_kanaal` "
                "attribute, or override the `get_kanaal()` method."
                % self.__class__.__name__
            )
        return self.notifications_kanaal

    def get_main_resource_key(self, kanaal: Kanaal) -> str:
        """
        Determine the key in the (response) data that represents the main resource.
        """
        if self.notifications_main_resource_key:
            return self.notifications_main_resource_key

        return kanaal.main_resource._meta.model_name

    def get_notification_main_object_url(self, data: dict, kanaal: Kanaal) -> str:
        """
        Retrieve the URL for the main object.
        """
        # using the main resource name, look up what the URL to this
        # object is/should be
        return data[self.get_main_resource_key(kanaal)]

    def construct_message(self, data: dict, instance: models.Model = None) -> dict:
        """
        Construct the message to send to the notification component.

        Using the response data from the view/action, we introspect this data
        to send the appropriate response. By convention, every resource
        includes its own absolute url in the 'url' key - we can use this to
        look up the object it points to. By convention, relations use the name
        of the resource, so for sub-resources we can use this to get a
        reference back to the main resource.
        """
        kanaal = self.get_kanaal()
        assert isinstance(kanaal, Kanaal), "`kanaal` should be a `Kanaal` instance"

        model = self.get_queryset().model

        if model is kanaal.main_resource:
            # look up the object in the database from its absolute URL
            resource_path = urlparse(data["url"]).path
            resource = instance or get_resource_for_path(resource_path)

            main_object = resource
            main_object_url = data["url"]
            main_object_data = data
        else:
            # lookup the main object from the URL
            main_object_url = self.get_notification_main_object_url(data, kanaal)
            main_object_path = urlparse(main_object_url).path
            main_object = get_resource_for_path(main_object_path)

            # get main_object data formatted by serializer
            view = get_viewset_for_path(main_object_path)
            serializer_class = view.get_serializer_class()
            serializer = serializer_class(
                main_object, context={"request": self.request}
            )
            main_object_data = serializer.data

        message_data = {
            "kanaal": kanaal.label,
            "hoofd_object": main_object_url,
            "resource": model._meta.model_name,
            "resource_url": data["url"],
            "actie": self.action,
            "aanmaakdatum": timezone.now(),
            # each channel knows which kenmerken it has, so delegate this
            "kenmerken": kanaal.get_kenmerken(main_object, main_object_data),
        }

        # let the serializer & render machinery shape the data the way it
        # should be, suitable for JSON in/output
        serializer = NotificatieSerializer(instance=message_data)
        return camelize(serializer.data)

    def notify(
        self, status_code: int, data: Union[List, Dict], instance: models.Model = None
    ) -> None:
        if get_setting("NOTIFICATIONS_DISABLED"):
            return

        # do nothing unless we have a 'success' status code - early exit here
        if not 200 <= status_code < 300:
            logger.info(
                "Not notifying, status code '%s' does not represent success.",
                status_code,
            )
            return

        # build the content of the notification
        message = self.construct_message(data, instance=instance)

        # build the client from the singleton config. This will raise an
        # exception if the config is not complete. We want this to hard-fail!
        client = NotificationsConfig.get_client()
        if client is None:
            raise RuntimeError("Could not build a client for Notifications API")

        # We've performed all the work that can raise uncaught exceptions that we can
        # still put inside an atomic transaction block. Next, we schedule the actual
        # sending block, which allows failures that are logged. Any unexpected errors
        # here will still cause the transaction to be committed (in the default
        # behaviour), but the exception will be visible in the error monitoring (such
        # as Sentry).
        #
        # The 'send_notification' task is passed down to the task queue on transaction commit

        def _send():
            send_notification.delay(message)

        transaction.on_commit(_send)


class NotificationCreateMixin(NotificationMixin):
    def create(self, request, *args, **kwargs):
        with conditional_atomic(self.notifications_wrap_in_atomic_block)():
            response = super().create(request, *args, **kwargs)
            self.notify(response.status_code, response.data)
            return response


class NotificationUpdateMixin(NotificationMixin):
    def update(self, request, *args, **kwargs):
        with conditional_atomic(self.notifications_wrap_in_atomic_block)():
            response = super().update(request, *args, **kwargs)
            self.notify(response.status_code, response.data)
            return response


class NotificationDestroyMixin(NotificationMixin):
    def destroy(self, request, *args, **kwargs):
        with conditional_atomic(self.notifications_wrap_in_atomic_block)():
            # get data via serializer
            instance = self.get_object()
            data = self.get_serializer(instance).data

            response = super().destroy(request, *args, **kwargs)
            self.notify(response.status_code, data, instance=instance)
            return response


class NotificationViewSetMixin(
    NotificationCreateMixin, NotificationUpdateMixin, NotificationDestroyMixin
):
    pass



import logging

import requests
from celery import shared_task
from zds_client import ClientError

from .autoretry import add_autoretry_behaviour
from .models import NotificationsConfig

logger = logging.getLogger(__name__)


class NotificationException(Exception):
    pass


@shared_task(bind=True)
def send_notification(self, message: dict) -> None:
    """
    send message to Notification API
    """
    client = NotificationsConfig.get_client()
    if client is None:
        logger.warning(
            "Could not build a client for Notifications API, not sending messages"
        )
        return

    try:
        client.create("notificaties", message)
    # any unexpected errors should show up in error-monitoring, so we only
    # catch ClientError exceptions
    except ClientError as exc:
        logger.warning(
            "Could not deliver message to %s",
            client.api_root,
            exc_info=exc,
            extra={
                "notification_msg": message,
                "current_try": self.request.retries + 1,
                "final_try": self.request.retries >= self.max_retries,
            },
        )

        raise NotificationException from exc


add_autoretry_behaviour(
    send_notification,
    autoretry_for=(
        NotificationException,
        requests.RequestException,
    ),
    retry_jitter=False,
)



# Generated by Django 3.2.15 on 2022-12-13 02:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("notifications_api_common", "0005_auto_20220810_0950"),
    ]

    operations = [
        migrations.AlterField(
            model_name="notificationsconfig",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
        migrations.AlterField(
            model_name="subscription",
            name="id",
            field=models.BigAutoField(
                auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
            ),
        ),
    ]



from threading import Thread

from django.db import transaction

from notifications_api_common.viewsets import NotificationMixin


def mock_notify(*args, **kwargs) -> None:
    def _send():
        NotificationMixin.send_notification(*args, **kwargs)

    transaction.on_commit(_send)


def mock_notify_async(*args, **kwargs) -> None:
    def _send():
        NotificationMixin.send_notification(*args, **kwargs)

    thread = Thread(target=_send)
    thread.start()
    thread.join()



# Generated by Django 3.2.14 on 2022-07-04 13:34

import django.contrib.postgres.fields
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="NotificationsConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "api_root",
                    models.URLField(
                        default="https://notificaties-api.vng.cloud/api/v1/",
                        unique=True,
                        verbose_name="api root",
                    ),
                ),
            ],
            options={
                "verbose_name": "Notificatiescomponentconfiguratie",
            },
        ),
        migrations.CreateModel(
            name="Subscription",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "callback_url",
                    models.URLField(
                        help_text="Where to send the notifications (webhook url)",
                        verbose_name="callback url",
                    ),
                ),
                (
                    "client_id",
                    models.CharField(
                        help_text="Client ID to construct the auth token",
                        max_length=50,
                        verbose_name="client ID",
                    ),
                ),
                (
                    "secret",
                    models.CharField(
                        help_text="Secret to construct the auth token",
                        max_length=50,
                        verbose_name="client secret",
                    ),
                ),
                (
                    "channels",
                    django.contrib.postgres.fields.ArrayField(
                        base_field=models.CharField(max_length=100),
                        help_text="Comma-separated list of channels to subscribe to",
                        size=None,
                        verbose_name="channels",
                    ),
                ),
                (
                    "_subscription",
                    models.URLField(
                        blank=True,
                        editable=False,
                        help_text="Subscription as it is known in the NC",
                        verbose_name="NC subscription",
                    ),
                ),
                (
                    "config",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="notifications_api_common.notificationsconfig",
                    ),
                ),
            ],
            options={
                "verbose_name": "Webhook subscription",
                "verbose_name_plural": "Webhook subscriptions",
            },
        ),
    ]



import sys
from typing import Any

from django.conf import settings

NOTIFICATIONS_DISABLED = False

IS_HTTPS = True


def get_setting(name: str) -> Any:
    this_module = sys.modules[__name__]
    default = getattr(this_module, name)
    return getattr(settings, name, default)
